@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2017 ECMWF.
+ * Copyright 2005-2018 ECMWF.
  *
  * This software is licensed under the terms of the Apache Licence Version 2.0
  * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
@@ -220,6 +220,7 @@ static int pack_long(grib_accessor* a, const long* val, size_t *len)
     case 1:  /* MARS labelling */
     case 36: /* MARS labelling for long window 4Dvar system */
     case 40: /* MARS labeling with domain and model (for LAM) */
+    case 42: /* LC-WFV: Wave forecast verification */
         if (isInstant) {
             /* type=em || type=es  */
             if (type==17) {
@@ -263,6 +264,7 @@ static int pack_long(grib_accessor* a, const long* val, size_t *len)
         }
         break;
 
+    case 5: /* Forecast probability data */
     case 7:  /* Sensitivity data */
     case 9:  /* Singular vectors and ensemble perturbations */
     case 11: /* Supplementary data used by the analysis */
