@@ -9,6 +9,7 @@
  */
 
 #include "grib_api_internal.h"
+#include <float.h>
 
 #ifdef GRIB_ON_WINDOWS
  /* Replace C99/Unix rint() for Windows Visual C++ (only before VC++ 2013 versions) */
@@ -31,6 +32,8 @@
  #endif
 #endif
 
+typedef enum {eROUND_ANGLE_UP, eROUND_ANGLE_DOWN} RoundingPolicy;
+
 static void set_total_length(unsigned char* buffer,long *section_length,long *section_offset,int edition,size_t totalLength)
 {
     long off;
@@ -98,10 +101,10 @@ static grib_handle* grib_sections_copy_internal(grib_handle* hfrom,grib_handle*
     p=buffer;
     off=0;
     for (i=0;i<=hfrom->sections_count;i++) {
-        grib_handle* h;
-        if (sections[i]) h=hfrom;
-        else h=hto;
-        p=(unsigned char*)memcpy(p,h->buffer->data+section_offset[i],section_length[i]);
+        grib_handle* hand = NULL;
+        if (sections[i]) hand=hfrom;
+        else hand=hto;
+        p=(unsigned char*)memcpy(p,hand->buffer->data+section_offset[i],section_length[i]);
         section_offset[i]=off;
         off+=section_length[i];
         p+=section_length[i];
@@ -334,9 +337,14 @@ static grib_trie* init_list(const char* name)
     return 0;
 }
 
-static void print_values(grib_context* c, const grib_util_grid_spec2* spec, const double* data_values,size_t data_values_count,const grib_values *values,int count)
+/* For debugging purposes */
+static void print_values(grib_context* c, const grib_util_grid_spec2* spec,
+        const double* data_values, const size_t data_values_count, /* the data pay load */
+        const grib_values *values, const size_t count) /* keys and their values */
 {
-    int i;
+    size_t i=0;
+    int isConstant = 1;
+    double v = 0, minVal=DBL_MAX, maxVal=-DBL_MAX;
     printf("GRIB_API DEBUG grib_util grib_set_values: setting %d values \n",count);
 
     for(i = 0; i < count ; i++)
@@ -352,29 +360,55 @@ static void print_values(grib_context* c, const grib_util_grid_spec2* spec, cons
         }
     }
 
+    printf("GRIB_API DEBUG  grib_util: data_values_count=%lu;\n", (unsigned long)data_values_count);
+    for (i=0; i<data_values_count; i++) {
+        if (i==0) v = data_values[i];
+        if (data_values[i] != spec->missingValue) {
+            if (v == spec->missingValue) {
+                v = data_values[i];
+            } else if (v != data_values[i]) {
+                isConstant=0;
+                break;
+            }
+        }
+    }
+
+    for (i=0; i<data_values_count; i++) {
+        v = data_values[i];
+        if (v!=spec->missingValue) {
+            if (v < minVal) minVal=v;
+            if (v > maxVal) maxVal=v;
+        }
+    }
+    printf("GRIB_API DEBUG  grib_util: data_values are CONSTANT? %d\t(min=%.16e, max=%.16e)\n",
+            isConstant, minVal, maxVal);
+
+#if 0
     if(spec->bitmapPresent) {
         int missing = 0;
+        size_t j = 0;
         double min = 1e100;
-        for(i = 0; i < data_values_count ; i++)
+        for(j = 0; j < data_values_count ; j++)
         {
-            double d = data_values[i] - spec->missingValue;
+            double d = data_values[j] - spec->missingValue;
             if(d < 0) d = -d;
-
             if(d < min) {
                 min = d;
             }
-
-            if(data_values[i] == spec->missingValue)
+            if(data_values[j] == spec->missingValue)
                 missing++;
         }
     }
+#endif
 }
 
+/*
 static int DBL_EQUAL(double d1, double d2, double tolerance)
 {
     return fabs(d1-d2) < tolerance;
 }
-
+*/
+#if 0
 /* Returns a boolean: 1 if angle can be encoded, 0 otherwise */
 static int grib1_angle_can_be_encoded(const double angle)
 {
@@ -387,8 +421,119 @@ static int grib1_angle_can_be_encoded(const double angle)
     return 0; /* sub millidegree. Cannot be encoded in grib1 */
 }
 
+/* Returns a boolean: 1 if angle can be encoded, 0 otherwise */
+static int angle_can_be_encoded(const double angle, const double angular_precision)
+{
+    const double angle_expanded = angle * angular_precision;
+    Assert(angular_precision>0);
+    double rounded = (long)(angle_expanded+0.5)/angular_precision;
+    if (angle<0) {
+        rounded = (long)(angle_expanded-0.5)/angular_precision;
+    }
+    if (angle == rounded) return 1;
+    /*printf("      ......... angle cannot be encoded: %.10e\n", angle);*/
+    return 0; /* Cannot be encoded */
+}
+#endif
+
+/* Returns a boolean: 1 if angle can be encoded, 0 otherwise */
+static int angle_can_be_encoded(grib_handle* h, const double angle)
+{
+    int ret = 0;
+    int retval = 1;
+    grib_handle* h2 = NULL;
+    char sample_name[16] = {0,};
+    long angular_precision = 0; /* e.g. 1e3 for grib1 and 1e6 for grib2 */
+    long edition = 0, coded = 0;
+    double expanded, diff;
+
+    if((ret = grib_get_long(h,"edition",&edition)) != 0) return ret;
+    if((ret = grib_get_long(h, "angularPrecision", &angular_precision)) != 0) return ret;
+    Assert(angular_precision > 0);
+
+    sprintf(sample_name, "GRIB%ld", edition);
+    h2 = grib_handle_new_from_samples(0, sample_name);
+    if((ret = grib_set_double(h2, "latitudeOfFirstGridPointInDegrees", angle)) != 0) return ret;
+    if((ret = grib_get_long(h2,   "latitudeOfFirstGridPoint", &coded)) != 0) return ret;
+    grib_handle_delete(h2);
+
+    expanded = angle*angular_precision;
+    diff = fabs(expanded - coded);
+    if (diff < 1.0/angular_precision)
+        retval = 1;
+    else
+        retval = 0;
+
+    return retval;
+}
+
+#ifdef GRIB_ON_WINDOWS
+#define round(a) ( (a) >=0 ? ((a)+0.5) : ((a)-0.5) )
+#endif
+static double adjust_angle(const double angle, const RoundingPolicy policy, const double angular_precision)
+{
+    double result = 0;
+    Assert(angular_precision > 0);
+    result = angle * angular_precision;
+    if (policy == eROUND_ANGLE_UP)
+        result = round(result+0.5);
+    else
+        result = round(result-0.5);
+    result = result / angular_precision;
+    return result;
+}
+
+/* Search key=value array for:
+ *   latitudeOfFirstGridPointInDegrees
+ *   longitudeOfFirstGridPointInDegrees
+ *   latitudeOfLastGridPointInDegrees
+ *   longitudeOfLastGridPointInDegrees
+ * and change their values to expand the bounding box
+ */
+static int expand_bounding_box(grib_handle* h, grib_values *values, const size_t count)
+{
+    int ret = GRIB_SUCCESS;
+    size_t i=0;
+    double new_angle = 0;
+    RoundingPolicy roundingPolicy = eROUND_ANGLE_UP;
+    long angular_precision = 0; /* e.g. 1e3 for grib1 and 1e6 for grib2 */
+    if((ret = grib_get_long(h, "angularPrecision", &angular_precision)) != 0)
+        return ret;
+
+    for(i=0; i<count; i++) {
+        int is_angle = 0;
+        if (strcmp(values[i].name, "longitudeOfFirstGridPointInDegrees")==0) {
+            roundingPolicy = eROUND_ANGLE_DOWN;
+            is_angle = 1;
+        }
+        else if (strcmp(values[i].name, "longitudeOfLastGridPointInDegrees")==0) {
+            roundingPolicy = eROUND_ANGLE_UP;
+            is_angle = 1;
+        }
+        else if (strcmp(values[i].name, "latitudeOfFirstGridPointInDegrees")==0) {
+            roundingPolicy = eROUND_ANGLE_UP;
+            is_angle = 1;
+        }
+        else if (strcmp(values[i].name, "latitudeOfLastGridPointInDegrees")==0) {
+            roundingPolicy = eROUND_ANGLE_DOWN;
+            is_angle = 1;
+        }
+
+        if (is_angle && !angle_can_be_encoded(h, values[i].double_value)) {
+            new_angle = adjust_angle(values[i].double_value, roundingPolicy, angular_precision);
+            if (h->context->debug) {
+                printf("GRIB_API DEBUG  grib_util EXPAND_BOUNDING_BOX %s: old=%.15e new=%.15e (%s)\n",
+                       values[i].name, values[i].double_value, new_angle,
+                       (roundingPolicy==eROUND_ANGLE_UP?"Up":"Down"));
+            }
+            values[i].double_value = new_angle;
+        }
+    }
+    return ret;
+}
+
 /* Returns a boolean: 1 if angle is too small, 0 otherwise */
-static int angle_too_small(const double angle, const double angular_precision)
+/*static int angle_too_small(const double angle, const double angular_precision)
 {
     const double a = fabs(angle);
     if (a > 0 && a < angular_precision) return 1;
@@ -400,8 +545,9 @@ static double normalise_angle(double angle)
     while (angle<0)   angle += 360;
     while (angle>360) angle -= 360;
     return angle;
-}
+}*/
 
+#if 0
 /* Check what is coded in the handle is what is requested by the spec. */
 /* Return GRIB_SUCCESS if the geometry matches, otherwise the error code */
 static int check_handle_against_spec(grib_handle* handle, const long edition,
@@ -527,6 +673,7 @@ static int check_handle_against_spec(grib_handle* handle, const long edition,
     }
     return GRIB_SUCCESS;
 }
+#endif
 
 static const char* get_grid_type_name(const int spec_grid_type)
 {
@@ -560,6 +707,27 @@ static const char* get_grid_type_name(const int spec_grid_type)
     return NULL;
 }
 
+static int is_constant_field(const double missingValue, const double* data_values, size_t data_values_count)
+{
+    int ii=0;
+    int constant=1;
+    double value = missingValue;
+
+    for (ii=0;ii<data_values_count;ii++) {
+        if (data_values[ii]!=missingValue) {
+            if (value==missingValue) {
+                value=data_values[ii];
+            } else {
+                if (value!=data_values[ii]) {
+                    constant=0;
+                    break;
+                }
+            }
+        }
+    }
+    return constant;
+}
+
 grib_handle* grib_util_set_spec(grib_handle* h,
         const grib_util_grid_spec    *spec,
         const grib_util_packing_spec *packing_spec,
@@ -643,9 +811,11 @@ grib_handle* grib_util_set_spec2(grib_handle* h,
     double laplacianOperator;
     int packingTypeIsSet=0;
     int setSecondOrder=0;
+    int setJpegPacking=0;
     size_t slen=17;
     int grib1_high_resolution_fix = 0; /* boolean: See GRIB-863 */
     int global_grid = 0; /* boolean */
+    int expandBoundingBox = 0;
 
     static grib_util_packing_spec default_packing_spec = {0, };
     Assert(h);
@@ -886,8 +1056,8 @@ grib_handle* grib_util_set_spec2(grib_handle* h,
         SET_LONG_VALUE("ijDirectionIncrementGiven", 1);
         if (editionNumber == 1) {
             /* GRIB-863: GRIB1 cannot represent increments less than a millidegree */
-            if (!grib1_angle_can_be_encoded(spec->iDirectionIncrementInDegrees) ||
-                !grib1_angle_can_be_encoded(spec->jDirectionIncrementInDegrees))
+            if (!angle_can_be_encoded(h, spec->iDirectionIncrementInDegrees) ||
+                    !angle_can_be_encoded(h, spec->jDirectionIncrementInDegrees))
             {
                 grib1_high_resolution_fix = 1;
                 /* Set flag to compute the increments */
@@ -1094,46 +1264,57 @@ grib_handle* grib_util_set_spec2(grib_handle* h,
     }
 
     switch(packing_spec->accuracy) {
-        case GRIB_UTIL_ACCURACY_SAME_BITS_PER_VALUES_AS_INPUT:
-        {
-            long bitsPerValue = 0;
-            Assert(grib_get_long(h, "bitsPerValue", &bitsPerValue) == 0);
-            SET_LONG_VALUE("bitsPerValue", bitsPerValue);
-        }
+    case GRIB_UTIL_ACCURACY_SAME_BITS_PER_VALUES_AS_INPUT:
+    {
+        long bitsPerValue = 0;
+        Assert(grib_get_long(h, "bitsPerValue", &bitsPerValue) == 0);
+        SET_LONG_VALUE("bitsPerValue", bitsPerValue);
+    }
+    break;
+
+    case GRIB_UTIL_ACCURACY_USE_PROVIDED_BITS_PER_VALUES:
+        SET_LONG_VALUE("bitsPerValue", packing_spec->bitsPerValue);
         break;
 
-        case GRIB_UTIL_ACCURACY_USE_PROVIDED_BITS_PER_VALUES:
-            SET_LONG_VALUE("bitsPerValue", packing_spec->bitsPerValue);
-            break;
+    case GRIB_UTIL_ACCURACY_SAME_DECIMAL_SCALE_FACTOR_AS_INPUT:
+    {
+        long decimalScaleFactor = 0;
+        Assert(grib_get_long(h, "decimalScaleFactor", &decimalScaleFactor) == 0);
+        SET_LONG_VALUE("decimalScaleFactor", decimalScaleFactor);
+    }
+    break;
 
-        case GRIB_UTIL_ACCURACY_SAME_DECIMAL_SCALE_FACTOR_AS_INPUT:
-        {
-            long decimalScaleFactor = 0;
-            Assert(grib_get_long(h, "decimalScaleFactor", &decimalScaleFactor) == 0);
-            SET_LONG_VALUE("decimalScaleFactor", decimalScaleFactor);
-        }
+    case GRIB_UTIL_ACCURACY_USE_PROVIDED_DECIMAL_SCALE_FACTOR:
+        SET_LONG_VALUE("decimalScaleFactor", packing_spec->decimalScaleFactor);
         break;
 
-        case GRIB_UTIL_ACCURACY_USE_PROVIDED_DECIMAL_SCALE_FACTOR:
-            SET_LONG_VALUE("decimalScaleFactor", packing_spec->decimalScaleFactor);
-            break;
-
-        default:
-            printf("invalid packing_spec->accuracy = %ld\n",(long)packing_spec->accuracy);
-            abort();
-            break;
+    default:
+        printf("invalid packing_spec->accuracy = %ld\n",(long)packing_spec->accuracy);
+        abort();
+        break;
     }
 
     if (packing_spec->extra_settings_count) {
         for(i = 0; i < packing_spec->extra_settings_count; i++) {
             Assert(count < 1024);
-            values[count++] = packing_spec->extra_settings[i];
-            if (strcmp(packing_spec->extra_settings[i].name, "global")==0 &&
-                packing_spec->extra_settings[i].long_value == 1)
+            if (strcmp(packing_spec->extra_settings[i].name, "expandBoundingBox")==0)
             {
-                /* GRIB-922: Request is for a global grid. Setting this key will
-                 * calculate the lat/lon values. So the spec's lat/lon can be ignored */
-                global_grid = 1;
+                if (packing_spec->extra_settings[i].long_value == 1) {
+                    /* ECC-625: Request is for expansion of bounding box (sub-area).
+                     * This is also called the "snap-out" policy */
+                    expandBoundingBox = 1;
+                }
+            }
+            else
+            {
+                values[count++] = packing_spec->extra_settings[i];
+                if (strcmp(packing_spec->extra_settings[i].name, "global")==0 &&
+                        packing_spec->extra_settings[i].long_value == 1)
+                {
+                    /* GRIB-922: Request is for a global grid. Setting this key will
+                     * calculate the lat/lon values. So the spec's lat/lon can be ignored */
+                    global_grid = 1;
+                }
             }
         }
     }
@@ -1169,12 +1350,24 @@ grib_handle* grib_util_set_spec2(grib_handle* h,
         }
     }
 
-    if (h->context->debug==-1)
+    if (h->context->debug==-1) {
+        printf("GRIB_API DEBUG grib_util: global_grid = %d\n", global_grid);
+        printf("GRIB_API DEBUG grib_util: expandBoundingBox = %d\n", expandBoundingBox);
         print_values(h->context,spec,data_values,data_values_count,values,count);
+    }
+
+    /* Apply adjustments to bounding box if needed */
+    if (expandBoundingBox) {
+        if ((*err=expand_bounding_box(outh, values, count)) != 0)
+        {
+            fprintf(stderr,"SET_GRID_DATA_DESCRIPTION: Cannot expand bounding box: %s\n",grib_get_error_message(*err));
+            goto cleanup;
+        }
+    }
 
     if ((*err = grib_set_values(outh,values,count)) != 0)
     {
-        fprintf(stderr,"SET_GRID_DATA_DESCRIPTION: Cannot set values  %s\n",grib_get_error_message(*err));
+        fprintf(stderr,"SET_GRID_DATA_DESCRIPTION: Cannot set key values: %s\n",grib_get_error_message(*err));
 
         for(i = 0; i < count; i++)
             if(values[i].error)
@@ -1185,17 +1378,17 @@ grib_handle* grib_util_set_spec2(grib_handle* h,
     if ((*err = grib_set_double_array(outh,"values",data_values,data_values_count)) != 0)
     {
         FILE* ferror;
-        long i,count;
+        size_t ii,lcount;
         grib_context* c=grib_context_get_default();
 
         ferror=fopen("error.data","w");
-        count=0;
+        lcount=0;
         fprintf(ferror,"# data_values_count=%ld\n",(long)data_values_count);
         fprintf(ferror,"set values={ ");
-        for (i=0;i<data_values_count-1;i++) {
-            fprintf(ferror,"%g, ",data_values[i]);
-            if (count>10) {fprintf(ferror,"\n");count=0;}
-            count++;
+        for (ii=0;ii<data_values_count-1;ii++) {
+            fprintf(ferror,"%g, ",data_values[ii]);
+            if (lcount>10) {fprintf(ferror,"\n");lcount=0;}
+            lcount++;
         }
         fprintf(ferror,"%g }",data_values[data_values_count-1]);
         fclose(ferror);
@@ -1203,18 +1396,17 @@ grib_handle* grib_util_set_spec2(grib_handle* h,
             grib_write_message(outh,"error.grib","w");
         goto cleanup;
     }
+
     /* grib_write_message(outh,"h.grib","w"); */
     /* if the field is empty GRIBEX is packing as simple*/
-    /*
-	if (!strcmp(input_packing_type,"grid_simple_matrix")) {
+    /*	if (!strcmp(input_packing_type,"grid_simple_matrix")) {
 		long numberOfValues;
 		grib_get_long(outh,"numberOfValues",&numberOfValues);
 		if (numberOfValues==0)  {
 			slen=11;
 			grib_set_string(outh,"packingType","grid_simple",&slen);
 		}
-	}
-     */
+	}     */
 
     if (grib1_high_resolution_fix) {
         /* GRIB-863: must set increments to MISSING */
@@ -1291,8 +1483,10 @@ grib_handle* grib_util_set_spec2(grib_handle* h,
     if (packing_spec->editionNumber && packing_spec->editionNumber!=editionNumber)
         grib_set_long(outh,"edition", packing_spec->editionNumber);
 
+#if 0
     if ( (*err = check_handle_against_spec(outh, editionNumber, spec, global_grid)) != GRIB_SUCCESS)
     {
+
         grib_context* c=grib_context_get_default();
         fprintf(stderr,"GRIB_UTIL_SET_SPEC: Geometry check failed! %s\n", grib_get_error_message(*err));
         if (editionNumber == 1) {
@@ -1302,7 +1496,7 @@ grib_handle* grib_util_set_spec2(grib_handle* h,
             grib_write_message(outh,"error.grib","w");
         goto cleanup;
     }
-
+#endif
     if (h->context->debug==-1)
         printf("GRIB_API DEBUG: grib_util_set_spec end\n");
 
@@ -1392,6 +1586,7 @@ int grib_moments(grib_handle* h,double east,double north,double west,double sout
     grib_context_free(c,lat);
     grib_context_free(c,lon);
     grib_context_free(c,values);
+    (void)mass;
 
     return ret;
 }
@@ -1540,7 +1735,7 @@ int parse_keyval_string(char* grib_tool, char* arg, int values_required, int def
                     printf("%s error: no value provided for key \"%s\"\n", grib_tool, values[i].name);
                 else
                     printf("Error: no value provided for key \"%s\"\n", values[i].name);
-                exit(GRIB_INVALID_ARGUMENT);
+                return GRIB_INVALID_ARGUMENT;
             }
             set_value(&values[i],value,equal);
         }
@@ -1636,9 +1831,9 @@ int is_gaussian_global(
 
     /* Note: final gaussian latitude = -first latitude */
     if ( (fabs(lat1 - latitudes[0]) >= d ) ||
-         (fabs(lat2 + latitudes[0]) >= d ) ||
-         lon1 != 0                         ||
-         lon2_diff > angular_precision
+            (fabs(lat2 + latitudes[0]) >= d ) ||
+            lon1 != 0                         ||
+            lon2_diff > angular_precision
     )
     {
         global = 0; /* sub area */
