@@ -71,5 +71,11 @@ ${tools_dir}grib_set -s setLocalDefinition=1,localDefinitionNumber=42,lcwfvSuite
 grib_check_key_equals $temp 'mars.origin:s' 'lops'
 rm -f $temp
 
+# Extra key in Local Definition 16 for GRIB1. ECC-1088
+# ----------------------------------------------------
+sample_g1=$GRIB_SAMPLES_PATH/reduced_gg_pl_640_grib1.tmpl
+${tools_dir}/grib_set -s setLocalDefinition=1,localDefinitionNumber=16,numberOfForecastsInEnsemble=51 $sample_g1 $temp
+grib_check_key_equals $temp 'totalNumber' '51'
+
 rm -f local.log loc.grib1 loc.grib2 loc1.grib1 loc1.grib2 eps.grib1 eps.grib2
 
